from typing import List, Optional

from ase.geometry.cell import cell_to_cellpar
from ase import Atoms
import numpy as np

import ase.io
import dpdata
import os
import glob

from .util import LammpsData
from ai2_kit.core.log import get_logger

logger = get_logger(__name__)


def dpdata_read_cp2k_dplr_data(
    cp2k_dir: str,
    cp2k_output: str,
    wannier_file: str,
    type_map: List[str],
    sel_type: List[int],
    wannier_cutoff: float = 1.0,
    wannier_spread_file: Optional[str] = None,
    model_charge_map: Optional[List[int]] = None,
):
    """
    Gnereate dpdata from cp2k output and wannier file for DPLR

    :param cp2k_dir: the directory of cp2k output and wannier file
    :param cp2k_output: the cp2k output file
    :param wannier_file: the wannier file
    :param type_map: the type map of atom type, for example, [O,H]
    :param sel_type: the selected type of atom, for example, [0] means atom type 0, aka O is selected
    :param wannier_cutoff: the cutoff to allocate wannier centers around atoms
    :param wannier_spread_file: the wannier spread file, if provided, the spread data will be added to dp_sys
    :param model_charge_map: the charge map of wannier in model, for example, [-8]

    :return dp_sys: dpdata.LabeledSystem
        In addition to the common energy data, atomic_dipole data is added.
    """
    cp2k_output = os.path.join(cp2k_dir, cp2k_output)
    wannier_file = os.path.join(cp2k_dir, wannier_file)
    wannier_spread_file = (
        os.path.join(cp2k_dir, wannier_spread_file) if wannier_spread_file else None
    )
    dp_sys = dpdata.LabeledSystem(cp2k_output, fmt="cp2k/output")
    try:
        dp_sys = set_dplr_ext_from_cp2k_output(
            dp_sys,
            wannier_file,
            type_map,
            sel_type,
            wannier_cutoff,
            wannier_spread_file,
            model_charge_map,
        )
    except ValueError:
        dp_sys = None

    return dp_sys


def set_dplr_ext_from_cp2k_output(
    dp_sys: dpdata.LabeledSystem,
    wannier_file: str,
    type_map: List[str],
    sel_type: List[int],
    wannier_cutoff: float = 1.0,
    wannier_spread_file: Optional[str] = None,
    model_charge_map: Optional[List[int]] = None,
):
    wannier_atoms = ase.io.read(wannier_file)
    # assert np.all(wannier_atoms.symbols == "X"), (
    #     "%s should include Wannier centres only" % wannier_file
    # )
    mask = (wannier_atoms.symbols == "X")  # type: ignore
    wannier_atoms = wannier_atoms[mask]

    natoms = dp_sys.get_natoms()
    nframes = dp_sys.get_nframes()
    if nframes == 0:
        return None
    assert nframes == 1, "Only support one frame"

    # symbols = np.array(dp_sys.data["atom_names"])[dp_sys.data["atom_types"]]
    sel_ids = get_sel_ids(dp_sys, type_map, sel_type)
    cns_ref = get_ref_cns(dp_sys, type_map, sel_type, model_charge_map)
    atomic_dipole, extended_coords, wannier_spread = get_atomic_dipole(
        dp_sys,
        sel_ids,
        wannier_atoms,
        wannier_cutoff,
        wannier_spread_file,
        cns_ref,
    )
    atomic_dipole_reformat = np.zeros((nframes, natoms, 3))
    atomic_dipole_reformat[:, sel_ids] = atomic_dipole
    atomic_dipole = atomic_dipole_reformat
    dp_sys.data["atomic_dipole"] = atomic_dipole.reshape([nframes, natoms, 3])
    try:
        wannier_spread_reformat = np.zeros((nframes, natoms, 4))
        if len(wannier_spread) > 0:
            wannier_spread_reformat[:, sel_ids] = wannier_spread.reshape([nframes, -1, 4])
            wannier_spread = wannier_spread_reformat
            dp_sys.data["wannier_spread"] = np.reshape(
                wannier_spread, [nframes, natoms, 4]
            )
    except:
        pass
    return dp_sys


def get_sel_ids(dp_sys, type_map, sel_type):
    symbols = np.array(dp_sys.data["atom_names"])[dp_sys.data["atom_types"]]
    # sel_ids = [np.where(symbols == type_map[ii])[0] for ii in sel_type]
    sel_ids = np.where(np.isin(symbols, np.array(type_map)[sel_type]))[0]
    return sel_ids


def get_ref_cns(dp_sys, type_map, sel_type, model_charge_map):
    if model_charge_map is None:
        return None
    else:
        cns_ref = []
        symbols = np.array(dp_sys.data["atom_names"])[dp_sys.data["atom_types"]]
        for s in symbols:
            for ii, idx in enumerate(sel_type):
                if s == type_map[idx]:
                    cns_ref.append(model_charge_map[ii])
        cns_ref = np.array(cns_ref) / (-2)
        return np.array(cns_ref, dtype=int)


def get_atomic_dipole(
    dp_sys,
    sel_ids,
    wannier_atoms,
    wannier_cutoff=1.0,
    wannier_spread_file=None,
    cns_ref=None,
):
    from MDAnalysis.lib.distances import distance_array, minimize_vectors

    coords = dp_sys.data["coords"].reshape(-1, 3)
    cellpar = cell_to_cellpar(dp_sys.data["cells"].reshape(3, 3))
    extended_coords = coords.copy()

    if wannier_spread_file:
        full_wannier_spread = read_wannier_spread(wannier_spread_file)[:, -1]
    else:
        full_wannier_spread = None

    ref_coords = coords[sel_ids].reshape(-1, 3)
    e_coords = wannier_atoms.get_positions()
    dist_mat = distance_array(ref_coords, e_coords, box=cellpar)
    atomic_dipole = []
    wannier_spread = []
    mlwc_ids = []
    if cns_ref is None:
        cns_ref = [4] * len(sel_ids)
    for ii, dist_vec in enumerate(dist_mat):
        mask = dist_vec < wannier_cutoff
        cn = np.sum(mask)
        if cn != cns_ref[ii]:
            raise ValueError(f"wannier atoms {ii} has {cn} atoms in the cutoff range")
        mlwc_ids.append(np.where(mask)[0])
        wc_coord_rel = e_coords[mask] - ref_coords[ii]
        if full_wannier_spread is not None:
            wannier_spread.append(full_wannier_spread[mask])
        wc_coord_rel = minimize_vectors(wc_coord_rel, box=cellpar)
        _atomic_dipole = wc_coord_rel.mean(axis=0)
        atomic_dipole.append(_atomic_dipole)
        wc_coord = _atomic_dipole + ref_coords[ii]
        extended_coords = np.concatenate(
            (extended_coords, wc_coord.reshape(1, 3)), axis=0
        )
    mlwc_ids = np.concatenate(mlwc_ids)
    # exclude double counting
    assert len(np.unique(mlwc_ids)) == np.sum(cns_ref)
    atomic_dipole = np.reshape(atomic_dipole, (-1, 3))
    assert atomic_dipole.shape[0] == len(sel_ids)
    wannier_spread = np.concatenate(wannier_spread)
    return atomic_dipole, extended_coords, wannier_spread


def build_sel_type_assertion(sel_type, model_path: str, py_cmd="python"):
    return f'''{py_cmd} -c "from deepmd.infer import DeepDipole;dp = DeepDipole({repr(model_path)});assert{repr(sel_type)}==[t for t in dp.tselt]"'''


def dump_dplr_lammps_data(
    fp,
    atoms: Atoms,
    type_map: List[str],
    sel_type: List[int],
    sys_charge_map: List[float],
    model_charge_map: List[float],
):
    """
    dump atoms to LAMMPS data file for DPLR
    the naming convention of params follows DeepMD-kit's

    about dplr: https://docs.deepmodeling.com/projects/deepmd/en/master/model/dplr.html
    about fitting tensor: https://docs.deepmodeling.com/projects/deepmd/en/master/model/train-fitting-tensor.html

    :param fp: file pointer
    :param type_map: the type map of atom type, for example, [O,H]
    :param sel_type: the selected type of atom, for example, [0] means atom type 0, aka O is selected
    :param sys_charge_map: the charge map of atom in system, for example, [6, 1]
    :param model_charge_map: the charge map of atom in model, for example, [-8]
    """

    if len(type_map) != len(sys_charge_map):
        raise ValueError(
            f"type_map {type_map} and sys_charge_map {sys_charge_map} must have the same length"
        )

    if len(sel_type) != len(model_charge_map):
        raise ValueError(
            f"sel_type {sel_type} and model_charge_map {model_charge_map} must have the same length"
        )

    new_atoms = atoms.copy()
    vtypes = get_unused_symbols(type_map, len(sel_type))

    r_atom_ids = []
    v_atom_ids = []

    # add virtual atoms
    for atype, vtype in zip(sel_type, vtypes):
        # create virtual atoms from the original atom type
        v_atoms = atoms[atoms.symbols == type_map[atype]]
        v_atoms.set_chemical_symbols([vtype] * len(v_atoms))  # type: ignore

        # create bonds between atom and its virtual atoms
        r_atom_ids.extend(np.where(atoms.symbols == type_map[atype])[0])
        v_atom_ids.extend(np.arange(len(v_atoms)) + len(new_atoms))  # type: ignore
        # this should be last
        new_atoms += v_atoms

    # build charges
    charges = np.zeros(len(new_atoms))
    for atype, charge in zip(type_map + vtypes, sys_charge_map + model_charge_map):
        charges[new_atoms.symbols == atype] = charge
    # build bonds
    n_bonds = len(r_atom_ids)
    # note: the index of lammps start from 1
    bonds = np.array(
        [
            np.arange(n_bonds) + 1,  # bond id
            [1] * n_bonds,  # bond type
            np.array(r_atom_ids) + 1,  # bond left
            np.array(v_atom_ids) + 1,  # bond right
        ],
        dtype=int,
    ).T

    # write lammps data
    lmp_data = LammpsData(new_atoms)
    lmp_data.charges = charges
    lmp_data.bonds = bonds
    lmp_data.write(fp, atom_style="full", specorder=type_map + vtypes)


def get_unused_symbols(used_symbols: List[str], size: int):
    """
    Get unused symbols from ase

    :param used_symbols: symbols that are already used
    :param size: number of unused symbols to get
    """
    from ase.data import chemical_symbols

    unused_symbols = []
    for symbol in reversed(chemical_symbols):
        if symbol not in used_symbols:
            unused_symbols.append(symbol)
        if len(unused_symbols) == size:
            break
    return unused_symbols


def get_sel_type(model_path) -> List[int]:
    try:
        from deepmd.infer import DeepDipole
    except ImportError:
        raise ImportError("deepmd-kit is not installed")

    dp = DeepDipole(model_path)
    return [t for t in dp.tselt]


def read_wannier_spread(fname: str):
    """
    Read wannier spread file generated by cp2k

    Parameters
    ----------
    fname : str
        wannier_spread.out file name

    Returns
    -------
    wannier_spread : numpy array
        wannier spread data
    """
    # skip 1, 2, and the last lines
    # save the others as array
    with open(fname, "r", encoding="UTF-8") as f:
        lines = f.readlines()[2:-1]

    wannier = []
    for line in lines:
        # line to array
        line = line.strip().split()
        wannier.append([float(line[1]), float(line[2])])
    return np.array(wannier)


def dplr_v2_to_v3(data_path: str, sel_symbol: list):
    atomic_data_fnames = [
        "atomic_dipole.npy",
        "atomic_polarizability.npy",
        "wannier_spread.npy",
    ]
    for atomic_data_fname in atomic_data_fnames:
        fnames = glob.glob(
            os.path.join(data_path, "**", atomic_data_fname), recursive=True
        )
        for fname in fnames:
            type_map = np.loadtxt(
                os.path.join(os.path.dirname(fname), "../type_map.raw"),
                dtype=str,
            )
            atype = np.loadtxt(
                os.path.join(os.path.dirname(fname), "../type.raw"),
                dtype=int,
            )
            symbols = np.array(type_map)[atype]
            sel_ids = np.where(np.isin(symbols, sel_symbol))[0]
            n_atoms = len(atype)

            raw_data = np.load(fname)
            n_frames = raw_data.shape[0]
            raw_data = np.reshape(raw_data, [n_frames, len(sel_ids), -1])
            n_dim = raw_data.shape[2]

            full_data = np.zeros([n_frames, n_atoms, n_dim])
            full_data[:, sel_ids] = raw_data
            np.save(fname, full_data.reshape([n_frames, -1]))


def dplr_v3_to_v2(data_path: str, sel_symbol: list):
    atomic_data_fnames = [
        "atomic_dipole.npy",
        "atomic_polarizability.npy",
        "wannier_spread.npy",
    ]
    for atomic_data_fname in atomic_data_fnames:
        fnames = glob.glob(
            os.path.join(data_path, "**", atomic_data_fname), recursive=True
        )
        for fname in fnames:
            type_map = np.loadtxt(
                os.path.join(os.path.dirname(fname), "../type_map.raw"),
                dtype=str,
            )
            atype = np.loadtxt(
                os.path.join(os.path.dirname(fname), "../type.raw"),
                dtype=int,
            )
            symbols = np.array(type_map)[atype]
            sel_ids = np.where(np.isin(symbols, sel_symbol))[0]
            n_atoms = len(atype)

            raw_data = np.load(fname)
            n_frames = raw_data.shape[0]
            raw_data_reshape = raw_data.reshape([n_frames, n_atoms, -1])
            np.save(fname, raw_data_reshape[:, sel_ids].reshape([n_frames, -1]))



import pandas as pd
from typing import List, Union


def load_colvar_from_files(*files):
    """
    load plumed colvar from multiple files and concatenate into a single DataFrame
    """
    fragments = []
    for file in files:
        with open(file) as f:
            fragments.append(load_colvar(f))
    return pd.concat(fragments)


def load_colvar(fp, skip='#! FIELDS'):
    """
    load plumed colvar file as DataFrame

    :param fp: file pointer
    :param skip: skip the header line
    :return: DataFrame
    """

    line = next(fp)
    if not line.startswith(skip):
        raise ValueError(f'Invalid colvar file header: {line}')
    names = line[len(skip):].strip().split()
    return pd.read_csv(fp, delim_whitespace=True, names=names)


def get_cvs_bias_from_df(df, cv_names: Union[List[str], str], bias_name: str):
    """
    get collective variables and bias from pandas dataframe

    :param df: pandas dataframe
    :param cv_names: names of collective variable columns
    :param bias_name: name of bias column

    :return: cvs, bias in numpy array for gaussian_kde
    """
    if isinstance(cv_names, str):
        cv_names = [cv_names]

    if len(cv_names) == 1:
        cvs = df[cv_names[0]].values
    else:
        cvs = df[cv_names].values.T

    bias = df[bias_name].values
    return cvs, bias





from dpdata.unit import econvs
from typing import List, Optional

import numpy as np

import dpdata
import os
import re


from .dplr import set_dplr_ext_from_cp2k_output, get_sel_ids
from ai2_kit.core.log import get_logger

logger = get_logger(__name__)


def dpdata_read_cp2k_dpff_data(
    cp2k_dir: str,
    cp2k_output: str,
    wannier_file: str,
    type_map: List[str],
    sys_charge_map: List[int],
    model_charge_map: List[int],
    ewald_h: float,
    ewald_beta: float,
    ext_efield,
    sel_type: List[int],
    wannier_cutoff: float = 1.0,
    wannier_spread_file: Optional[str] = None,
):
    """
    Gnereate dpdata from cp2k output and wannier file for DPLR

    :param cp2k_dir: the directory of cp2k output and wannier file
    :param cp2k_output: the cp2k output file
    :param wannier_file: the wannier file
    :param type_map: the type map of atom type, for example, [O,H]
    :param sys_charge_map: the charge map of atom in system, for example, [6, 1]
    :param model_charge_map: the charge map of wannier in model, for example, [-8]
    :param ewald_h: the ewald_h parameter used in dplr/dpff model
    :param ewald_beta: the ewald_beta parameter used in dplr/dpff model
    :param ext_efield: the external electric field
    :param sel_type: the selected type of atom, for example, [0] means atom type 0, aka O is selected
    :param wannier_cutoff: the cutoff to allocate wannier centers around atoms
    :param wannier_spread_file: the wannier spread file, if provided, the spread data will be added to dp_sys

    :return dp_sys: dpdata.LabeledSystem
        In addition to the common energy data, the following data are added:
        - atomic_dipole: the atomic dipole of selected atoms
        - ext_efield: the external electric field
        - efield: the direction vector of atomic electric field
        - aparam: the atomic electric field magnitude
    """
    cp2k_output = os.path.join(cp2k_dir, cp2k_output)
    wannier_file = os.path.join(cp2k_dir, wannier_file)
    wannier_spread_file = (
        os.path.join(cp2k_dir, wannier_spread_file) if wannier_spread_file else None
    )
    dp_sys = dpdata.LabeledSystem(cp2k_output, fmt="cp2k/output")
    try:
        dp_sys = set_dpff_ext_from_cp2k_output(
            dp_sys,
            cp2k_output,
            wannier_file,
            type_map,
            sys_charge_map,
            model_charge_map,
            ewald_h,
            ewald_beta,
            ext_efield,
            sel_type,
            wannier_cutoff,
            wannier_spread_file,
        )
    except ValueError:
        dp_sys = None

    return dp_sys


def set_dpff_ext_from_cp2k_output(
    dp_sys: dpdata.LabeledSystem,
    cp2k_output: str,
    wannier_file: str,
    type_map: List[str],
    sys_charge_map: List[int],
    model_charge_map: List[int],
    ewald_h: float,
    ewald_beta: float,
    ext_efield,
    sel_type: List[int],
    wannier_cutoff: float = 1.0,
    wannier_spread_file: Optional[str] = None,
):
    # with atomic_dipole
    # update energy
    # add ext_efield, efield, aparam
    dplr_dp_sys = set_dplr_ext_from_cp2k_output(
        dp_sys,
        wannier_file,
        type_map,
        sel_type,
        wannier_cutoff,
        wannier_spread_file,
        model_charge_map,
    )
    atomic_dipole = dplr_dp_sys.data["atomic_dipole"].reshape(-1, 3)  # type: ignore

    with open(cp2k_output, "r", encoding="UTF-8") as fp:
        raw_energy = get_cp2k_output_total_energy(fp)

    ext_efield = np.reshape(ext_efield, [1, 3])
    natoms = dp_sys.get_natoms()
    nframes = dp_sys.get_nframes()
    if nframes == 0:
        return None
    assert nframes == 1, "Only support one frame"

    symbols = np.array(dp_sys.data["atom_names"])[dp_sys.data["atom_types"]]
    sel_ids = get_sel_ids(dp_sys, type_map, sel_type)

    ion_coords = dp_sys.data["coords"].reshape(-1, 3)
    # atomic_dipole.shape = (natoms * 3)
    wannier_coords = (ion_coords + atomic_dipole)[sel_ids].reshape(-1, 3)
    extended_coords = np.concatenate((ion_coords, wannier_coords), axis=0)

    # get extended charges
    extended_charges = np.zeros(len(extended_coords))
    for symbol, ion_charge in zip(type_map, sys_charge_map):
        extended_charges[np.where(symbols == symbol)[0]] = ion_charge
    sel_symbols = symbols[sel_ids]
    for ii, wannier_charge in zip(sel_type, model_charge_map):
        sel_symbol = type_map[ii]
        extended_charges[np.where(sel_symbols == sel_symbol)[0] + natoms] = (
            wannier_charge
        )
    total_dipole = np.sum(extended_coords * extended_charges.reshape(-1, 1), axis=0)
    energy = raw_energy - np.inner(total_dipole, ext_efield.reshape(3))
    dp_sys.data["energies"] = np.reshape(energy, [-1])
    pbc_efield = get_pbc_atomic_efield(
        dp_sys, extended_coords, extended_charges, ewald_h, ewald_beta
    )
    # natoms * 3
    _efield = pbc_efield + ext_efield
    aparam = np.linalg.norm(_efield, axis=-1)
    efield = np.array(_efield) / aparam.reshape(natoms, 1)

    dp_sys.data["ext_efield"] = ext_efield.reshape(nframes, 3)
    dp_sys.data["efield"] = efield.reshape(nframes, natoms, 3)
    dp_sys.data["aparam"] = aparam.reshape(nframes, natoms, -1)
    return dp_sys


def get_cp2k_output_total_energy(fp):
    start_pattern = r"  Total charge density g-space grids:"
    end_pattern = r"  Total energy:"
    start_pattern = re.compile(start_pattern)
    end_pattern = re.compile(end_pattern)

    flag = False
    data_lines = []
    nloop = 0
    for line in fp:
        line = line.strip("\n")
        if start_pattern.match(line):
            flag = True
        elif flag and end_pattern.match(line):
            flag = False
            nloop += 1

        if flag is True:
            data_lines.append(line)
    data_lines = np.reshape(data_lines, (nloop, -1))

    tot_e = 0.0
    # grep the data from the last iteration
    for kw in data_lines[-1, 2:-1]:
        kw = kw.split(":")
        k = kw[0].strip(" ")
        v = float(kw[1]) * econvs["hartree"]
        if k != "Fermi energy":
            tot_e += v
    return tot_e


def get_pbc_atomic_efield(
    dp_sys,
    extended_coords,
    extended_charges,
    ewald_h,
    ewald_beta,
):
    from deepmd.infer.ewald_recp import EwaldRecp

    er = EwaldRecp(ewald_h, ewald_beta)
    natoms = dp_sys.get_natoms()
    cell = dp_sys.data["cells"]
    e, f, v = er.eval(
        extended_coords.reshape(1, -1),
        extended_charges.reshape(1, -1),
        cell.reshape(1, -1),
    )
    extended_efield = f.reshape(-1, 3) / extended_charges.reshape(-1, 1)
    # natoms * 3
    pbc_efield = extended_efield[:natoms]
    return pbc_efield



from ai2_kit.feat.spectrum.viber import dpdata_read_cp2k_viber_data
from ai2_kit.domain.dplr import dpdata_read_cp2k_dplr_data, dplr_v3_to_v2
from ai2_kit.core.util import (
    ensure_dir,
    expand_globs,
    list_sample,
    SAMPLE_METHOD,
    slice_from_str,
)
from ai2_kit.core.log import get_logger

import os
import glob

from typing import Optional
from dpdata.data_type import Axis, DataType
import numpy as np
import dpdata


logger = get_logger(__name__)


def register_data_types():
    if getattr(dpdata, "__registed__", False):
        return

    DATA_TYPES = [
        DataType("fparam", np.ndarray, (Axis.NFRAMES, -1), required=False),  # type: ignore
        DataType("aparam", np.ndarray, (Axis.NFRAMES, Axis.NATOMS, -1), required=False),  # type: ignore
        DataType("efield", np.ndarray, (Axis.NFRAMES, Axis.NATOMS, 3), required=False),  # type: ignore
        DataType("ext_efield", np.ndarray, (Axis.NFRAMES, 3), required=False),  # type: ignore
        DataType("atomic_dipole", np.ndarray, (Axis.NFRAMES, Axis.NATOMS, 3), required=False),  # type: ignore
        DataType("atomic_polarizability", np.ndarray, (Axis.NFRAMES, Axis.NATOMS, 9), required=False),  # type: ignore
        DataType("wannier_spread", np.ndarray, (Axis.NFRAMES, Axis.NATOMS, 4), required=False),  # type: ignore
    ]
    dpdata.System.register_data_type(*DATA_TYPES)  # type: ignore
    dpdata.LabeledSystem.register_data_type(*DATA_TYPES)  # type: ignore
    dpdata.__registed__ = True  # type: ignore


register_data_types()


class DpdataTool:

    def __init__(self, verbose=False, systems: Optional[list] = None):
        self._systems = [] if systems is None else systems
        self._verbose = verbose

    def read(self, *file_path_or_glob: str, **kwargs):
        """
        read data from multiple paths, support glob pattern
        default format is deepmd/npy

        :param file_path_or_glob: path or glob pattern to locate data path
        :param fmt: format to read, default is deepmd/npy
        :param label: default is True, use dpdata.LabeledSystem if True, else use dpdata.System
        :param kwargs: arguments to pass to dpdata.System or dpdata.LabeledSystem
        """
        systems = read(*file_path_or_glob, **kwargs)
        self._systems.extend(systems)
        return self

    def filter(self, lambda_expr: str):
        """
        filter data with lambda expression

        :param lambda_expr: lambda expression to filter data
        """
        fn = eval(lambda_expr)
        self._systems = [system for system in self._systems if fn(system.data)]
        return self

    def slice(self, expr: str):
        """
        slice systems by python slice expression, for example
        `10:`, `:10`, `::2`, etc

        :param start: start index
        :param stop: stop index
        :param step: step
        """
        s = slice_from_str(expr)
        self._systems = self._systems[s]
        return self

    def sample(self, size: int, method: SAMPLE_METHOD = "even", **kwargs):
        """
        sample data

        :param size: size of sample, if size is larger than data size, return all data
        :param method: method to sample, can be 'even', 'random', 'truncate', default is 'even'
        :param seed: seed for random sample, only used when method is 'random'

        Note that by default the seed is length of input list,
        if you want to generate different sample each time, you should set random seed manually
        """
        self._systems = list_sample(self._systems, size, method, **kwargs)
        return self

    def size(self):
        """
        size of loaded data
        """
        print(len(self._systems))
        return self

    def write(
        self,
        out_path: str,
        fmt="deepmd/npy",
        merge: bool = True,
        v2: bool = False,
        sel_symbol: Optional[list] = None,
    ):
        """
        write data to specific path, support deepmd/npy, deepmd/raw, deepmd/hdf5 formats
        :param out_path: path to write data
        :param fmt: format to write, default is deepmd/npy
        :param merge: if True, merge all data use dpdata.MultiSystems, else write data without merging
        :param v2: if True, write data in v2 format, else write data in v3 format
        :param sel_symbol: the selected symbols of atom, for example, ["O", "K", "F"] means only write data of O, K, F
        """
        ensure_dir(out_path)
        if len(self._systems) == 0:
            raise ValueError("No data to merge")
        if merge:
            systems = dpdata.MultiSystems(self._systems[0])
        else:
            systems = self._systems[0]

        for system in self._systems[1:]:
            systems.append(system)

        if fmt == "deepmd/npy":
            systems.to_deepmd_npy(out_path)  # type: ignore
            if v2:
                assert (
                    sel_symbol is not None
                ), "sel_symbol must be provided when v2 is True"
                dplr_v3_to_v2(out_path, sel_symbol)

        elif fmt == "deepmd/raw":
            systems.to_deepmd_raw(out_path)  # type: ignore
        elif fmt == "deepmd/hdf5":
            systems.to_deepmd_hdf5(out_path)  # type: ignore
        else:
            raise ValueError(f"Unknown fmt {fmt}")

    def set_fparam(self, fparam):
        """
        Set fparam for all systems

        :param fparam: fparam to set, should be a scalar or vector, e.g. 1.0 or [1.0, 2.0]
        """
        for system in self._systems:
            set_fparam(system, fparam)
        return self

    def eval(self, dp_model: str):
        """
        Use deepmd model to label energy, force and viral

        :param dp_model: path to deepmd frozen model
        """
        from deepmd.infer import DeepPot

        systems = dpdata.System()
        systems.extend(self._systems)  # merge systems to one

        pot = DeepPot(dp_model, auto_batch_size=True)  # type: ignore

        # remap atypes to pot's type
        atom_names = systems.get_atom_names()
        target_atom_names = pot.get_type_map()
        mapping = {
            i: target_atom_names.index(name) for i, name in enumerate(atom_names)
        }
        vectorized_mapping = np.vectorize(mapping.get)

        atypes = vectorized_mapping(systems.data["atom_types"])
        coords = systems.data["coords"]
        cells = None if systems.nopbc else systems.data["cells"]

        e, f, v = pot.eval(coords=coords, cells=cells, atom_types=atypes)  # type: ignore

        n_atoms = systems.get_natoms()
        n_frames = systems.get_nframes()

        e = e.reshape((n_frames,))
        f = f.reshape((n_frames, n_atoms, 3))
        v = v.reshape((n_frames, 3, 3))

        data = {**systems.data, "energies": e, "forces": f, "virials": v}
        # replace system files
        self._systems = []
        self._systems.extend(dpdata.LabeledSystem.from_dict({"data": data}))  # type: ignore
        return self

    def to_ase(self):
        """
        Convert dpdata format to ase format, and use ase tool to handle
        """
        from .ase import AseTool

        atoms_list = []
        for sys in self._systems:
            atoms_list.extend(sys.to_ase_structure())
        return AseTool(atoms_arr=atoms_list)

    def _verbose_log(self, msg, **kwargs):
        if self._verbose:
            logger.info(msg, **kwargs)


def set_fparam(system, fparam):
    nframes = system.get_nframes()
    system.data["fparam"] = np.tile(fparam, (nframes, 1))
    return system


def read(*file_path_or_glob: str, **kwargs):
    """
    read data from multiple paths, support glob pattern
    default format is deepmd/npy

    :param file_path_or_glob: path or glob pattern to locate data path
    :param fmt: format to read, default is deepmd/npy
    :param label: default is True, use dpdata.LabeledSystem if True, else use dpdata.System
    :parse ignore_error: if True, ignore error when read data, default is False
    :param kwargs: arguments to pass to dpdata.System or dpdata.LabeledSystem
    """
    kwargs.setdefault("fmt", "deepmd/npy")
    ignore_error = kwargs.pop("ignore_error", False)
    files = expand_globs(file_path_or_glob)
    if len(files) == 0:
        raise FileNotFoundError(f"No file found in {file_path_or_glob}")
    systems = []
    for file in files:
        try:
            system = _read(file, **kwargs)
            if system is not None:
                systems.extend(system)
        except Exception:
            if not ignore_error:
                raise
            logger.exception(f"Fail to process file {file}, ignore and continue")
    return systems


def _read(data_path: str, **kwargs):
    # pop custom arguments or else it will be passed to dpdata.System and raise error
    fmt = kwargs.pop("fmt", "deepmd/npy")
    fparam = kwargs.pop("fparam", None)
    label = kwargs.pop("label", True)

    if fmt == "cp2k/viber":
        system = dpdata_read_cp2k_viber_data(data_path, **kwargs)
    elif fmt == "cp2k/dplr":
        system = dpdata_read_cp2k_dplr_data(data_path, **kwargs)
    else:
        system = (
            dpdata.LabeledSystem(data_path, fmt=fmt, **kwargs)
            if label
            else dpdata.System(data_path, fmt=fmt, **kwargs)
        )

    if fparam is not None:
        set_fparam(system, fparam)

    return system



from typing import List, Tuple
from ase import Atoms

import pandas as pd

from ai2_kit.core.util import expand_globs
from ai2_kit.core.log import get_logger

logger = get_logger(__name__)

MdArray = List[Tuple[pd.DataFrame, str]]


class ModelDevi:
    """
    A tool to analyze the deviation of model from model deviation file of deepmd-kit
    """

    @staticmethod
    def md_arr_load(*files: str) -> MdArray:
        """
        Load model deviation files
        """
        md_arr = []
        for file in expand_globs(files):
            with open(file, 'r') as f:
                f.seek(1)  # skip the leading '#'
                df = pd.read_csv(f, delim_whitespace=True)
                md_arr.append((df, file))
        return md_arr

    def __init__(self, atoms_arr: List[Atoms], md_arr: MdArray):
        self._md_df = pd.concat([df for df, _ in md_arr])
        self._atoms_arr = atoms_arr
        self._md_arr = md_arr
        if len(atoms_arr) != len(self._md_df):
            raise ValueError("The size of atoms and model deviation records should be the same")
        self._stats = {}
        self._grade = {}

    def grade(self, lo: float, hi: float, col: str = 'max_devi_f'):
        """
        Grade atoms based on the deviation of model: the good, the bad and the ugly
        the grade is based on the column of max_devi_f by default,
        if the value is below lo, the level is good,
        if the value is above hi, the level is ugly,
        otherwise, the level is bad

        :param lo: the lower bound of good level
        :param hi: the upper bound of ugly level
        :param col: the column of model deviation to grade, default is max_devi_f
        """
        if col not in self._md_df.columns:
            raise ValueError(f"Unknown model deviation column: {col}")
        for df, file in self._md_arr:
            good = df[col] < lo
            bad = (df[col] >= lo) & (df[col] <= hi)
            ugly = df[col] > hi
            self._stats[file] = {
                'g': good.sum(),
                'b': bad.sum(),
                'u': ugly.sum(),
                'all': len(df),
            }
        self._grade['good'] = self._md_df[col] < lo
        self._grade['bad'] = (self._md_df[col] >= lo) & (self._md_df[col] <= hi)
        self._grade['ugly'] = self._md_df[col] > hi
        return self

    def dump_stats(self, out_file: str = '', fmt='tsv'):
        """
        Dump the statistics of grading
        """
        from tabulate import tabulate

        headers = ['file', 'total', 'good', 'bad', 'ugly', 'good%', 'bad%', 'ugly%']
        table = []

        overall = {
            'all': len(self._md_df),
            'g': self._grade['good'].sum(),
            'b': self._grade['bad'].sum(),
            'u': self._grade['ugly'].sum(),
        }
        for file, stats in [*self._stats.items(), ('', overall)]:
            total = stats['all']
            g = stats['g']
            b = stats['b']
            u = stats['u']
            g_pct = '{:.2%}'.format(g / total)
            b_pct = '{:.2%}'.format(b / total)
            u_pct = '{:.2%}'.format(u / total)
            table.append([file, total, g, b, u, g_pct, b_pct, u_pct])
        stats_report = tabulate(table, headers=headers, tablefmt=fmt)
        if out_file:
            with open(out_file, 'w') as f:
                f.write(stats_report)
        else:
            logger.info(f'model deviation statistics:\n{stats_report}')
        return self

    def to_ase(self, level):
        """
        Hand over the atoms to ase tool

        :param level: the grade level to hand over, valid values are good, bad, ugly
        """
        if level not in self._grade:
            raise ValueError(f"Unknown grade level: {level}")
        atoms_arr = [self._atoms_arr[i] for i in self._grade[level].index if self._grade[level][i]]
        from .ase import AseTool
        return AseTool(atoms_arr)



from ai2_kit.core.artifact import Artifact, ArtifactDict
from ai2_kit.core.script import BashScript, BashStep, BashTemplate
from ai2_kit.core.job import gather_jobs
from ai2_kit.core.util import dict_nested_get, dump_json, list_split, list_sample
from ai2_kit.core.log import get_logger
from ai2_kit.core.pydantic import BaseModel

from typing import List, Union, Optional, Tuple, Literal
from pymatgen.io.vasp.inputs import Incar, Kpoints
from dataclasses import dataclass

import ase.io
from ase import Atoms
from ase.io.vasp import _symbol_count_from_symbols

import copy
import os

from .data import DataFormat, artifacts_to_ase_atoms
from .iface import ICllLabelOutput, BaseCllContext

logger = get_logger(__name__)

class CllVaspInputConfig(BaseModel):
    init_system_files: List[str] = []
    input_template: Optional[Union[dict, str]] = None
    """
    INCAR template for VASP. Could be a dict or content of a VASP input file.
    """
    potcar_source: Optional[dict] = {}
    """
    POTCAR source for VASP. Could be a dict or list of paths.
    """
    kpoints_template: Optional[Union[dict, str]] = None
    """
    KPOINTS template for VASP. Could be a dict or content of a VASP input file.
    """
    limit: int = 50
    """
    Limit of the number of systems to be labeled.
    """
    limit_method: Literal["even", "random", "truncate"] = "even"

    ignore_error: bool = False
    """
    Ignore error when running VASP.
    """

class CllVaspContextConfig(BaseModel):
    script_template: BashTemplate
    vasp_cmd: str = 'vasp_std'
    post_vasp_cmd: str = 'echo "no post_vasp_cmd"'
    concurrency: int = 5

@dataclass
class CllVaspInput:
    config: CllVaspInputConfig
    system_files: List[Artifact]
    type_map: List[str]
    initiated: bool = False  # FIXME: this seems to be a bad design idea


@dataclass
class CllVaspContext(BaseCllContext):
    config: CllVaspContextConfig


@dataclass
class GenericVaspOutput(ICllLabelOutput):
    vasp_outputs: List[Artifact]

    def get_labeled_system_dataset(self):
        return self.vasp_outputs


async def cll_vasp(input: CllVaspInput, ctx: CllVaspContext) -> GenericVaspOutput:
    executor = ctx.resource_manager.default_executor

    # For the first round
    # FIXME: move out from this function, this should be done in the workflow
    if not input.initiated:
        input.system_files += ctx.resource_manager.resolve_artifacts(input.config.init_system_files)

    if len(input.system_files) == 0:
        return GenericVaspOutput(vasp_outputs=[])

    # setup workspace
    work_dir = os.path.join(executor.work_dir, ctx.path_prefix)
    [tasks_dir] = executor.setup_workspace(work_dir, ['tasks'])

    # create task dirs and prepare input files
    vasp_task_dirs = executor.run_python_fn(make_vasp_task_dirs)(
        system_files=[a.to_dict() for a in input.system_files],
        type_map=input.type_map,
        base_dir=tasks_dir,
        input_template=input.config.input_template,
        potcar_source=input.config.potcar_source,
        kpoints_template=input.config.kpoints_template,
        limit=input.config.limit,
        limit_method=input.config.limit_method,
    )

    # build commands
    steps = []
    for vasp_task_dir in vasp_task_dirs:
        cmd = f'{ctx.config.vasp_cmd} &> output && {ctx.config.post_vasp_cmd}'
        steps.append(BashStep(
            cwd=vasp_task_dir['url'],
            cmd=cmd,
            checkpoint='vasp',
            exit_on_error=not input.config.ignore_error,
        ))

    # submit tasks and wait for completion
    jobs = []
    for i, steps_group in enumerate(list_split(steps, ctx.config.concurrency)):
        if not steps_group:
            continue
        script = BashScript(
            template=ctx.config.script_template,
            steps=steps_group,
        )
        job = executor.submit(script.render(), cwd=tasks_dir)
        jobs.append(job)
    jobs = await gather_jobs(jobs, max_tries=2)

    vasp_outputs = [Artifact.of(
        url=a['url'],
        format=DataFormat.VASP_OUTPUT_DIR,
        executor=executor.name,
        attrs=a['attrs'],
    ) for a in vasp_task_dirs]

    return GenericVaspOutput(vasp_outputs=vasp_outputs)


def make_vasp_task_dirs(system_files: List[ArtifactDict],
                        type_map: List[str],
                        input_template: Optional[Union[dict, str]],
                        potcar_source: Optional[dict],
                        base_dir: str,
                        kpoints_template: Optional[Union[dict, str]] = None,
                        limit: int = 0,
                        limit_method: Literal["even", "random", "truncate"] = "even"
                        ) -> List[ArtifactDict]:
    """Generate VASP input files from LAMMPS dump files or XYZ files."""

    task_dirs = []
    atoms_list: List[Tuple[ArtifactDict, Atoms]] = artifacts_to_ase_atoms(system_files, type_map=type_map)

    if limit > 0:
        atoms_list = list_sample(atoms_list, limit, method=limit_method)

    for i, (data_file, atoms) in enumerate(atoms_list):
        # create task dir
        task_dir = os.path.join(base_dir, f'{str(i).zfill(6)}')
        os.makedirs(task_dir, exist_ok=True)
        dump_json(data_file, os.path.join(task_dir, 'debug.data-file.json'))

        # load system-wise config from attrs
        overridable_params: dict = copy.deepcopy(dict_nested_get(data_file, ['attrs', 'vasp'], dict()))  # type: ignore

        # create input file
        input_template = overridable_params.get('input_template', input_template)

        # prepare input template
        if isinstance(input_template, str):
            input_template = Incar.from_file(input_template).as_dict()

        assert input_template, 'input_template must be provided'
        incar = Incar.from_dict(input_template)
        incar.write_file(os.path.join(task_dir, 'INCAR'))

        # create POSCAR
        elements_all = atoms.get_chemical_symbols()
        elements = [
            item[0] for item in _symbol_count_from_symbols(elements_all)
        ]
        ase.io.write(
            os.path.join(task_dir, 'POSCAR'), atoms, format='vasp'
        )

        # create POTCAR
        potcar_source = overridable_params.get('potcar_source', potcar_source)

        # prepare potcar
        assert potcar_source, 'potcar_source must be provided'
        with open(os.path.join(task_dir, 'POTCAR'), 'w') as out_file:
            for element in elements:
                with open(potcar_source[element], 'r') as in_file:
                    out_file.write(in_file.read())

        # create KPOINTS
        kpoints_template = overridable_params.get('kpoints_template', kpoints_template)

        # prepare kpoints template
        if isinstance(kpoints_template, str):
            kpoints_template = Kpoints.from_file(kpoints_template).as_dict()
        
        if kpoints_template:
            kpoints = Kpoints.from_dict(kpoints_template)
            kpoints.write_file(os.path.join(task_dir, 'KPOINTS'))

        # inherit attrs from input file
        # TODO: inherit only ancestor key should be enough
        task_dirs.append({
            'url': task_dir,
            'attrs': data_file['attrs'],
        })

    return task_dirs
